import os
import re
import time
from typing import List

from .arpc_template import generate_param_class, generate_procedure_class


PACKAGE = 'python'


def generate_package(arpc_meta, path, output, async_):
    # mkdir if not exists
    if not os.path.exists(output):
        os.makedirs(output)

    file_name = os.path.basename(path)

    package_name = file_name.replace(".", "_")
    if len(arpc_meta['package']) > 0:
        for pkg in arpc_meta['package']:
            if pkg['language'] == PACKAGE:
                package_name = pkg['name']
                break

    python_file = os.path.join(output, file_name.replace(".arpc", ".py"))

    localtime = time.asctime(time.localtime(time.time()))

    file_str = f"""# -*- coding: utf-8 -*-
# Generated by arpc-python at {localtime}
# Path: {path}
# package: {package_name}

import abc
from arpc.base import {'BaseAsync' if async_ else 'Base'}
"""

    for k, v in arpc_meta['param'].items():
        result = generate_param_class(k, v, async_)
        file_str += result

    file_str += generate_procedure_class('Client',
                                         arpc_meta['unique'], arpc_meta['procedures'], async_)

    if os.path.exists(python_file):
        os.remove(python_file)

    if not os.path.exists(python_file):
        with open(python_file, "w") as f:
            f.write(file_str)

    return python_file


def compile_arpc(path: str) -> dict:
    if not os.path.exists(path):
        raise FileNotFoundError(f"File '{path}' does not exist")

    # if not path.endswith(".arpc"):
    #     raise FileNotFoundError(f"File '{path}' is not an .arpc file")

    arpc_meta = {}

    # 当前解析到的行号
    line_num = 0
    # 正在解析 package
    handle_package = False
    # 正在解析 procedure
    handle_procedures = False
    # 正在解析 param
    handle_param = False
    # 当前解析 param 名
    param_name = ''

    with open(path, 'r') as f:
        lines = f.readlines()
        for line in lines:
            line_num += 1
            # 去掉换行
            line = line.strip()
            # 去掉注释
            line = re.sub(r"//.*", "", line)
            # 去掉空行
            if not line:
                continue

            # 正在处理 package
            if handle_package:
                if line.startswith("}"):
                    handle_package = False
                    continue
                elif line.startswith("{"):
                    continue
                else:
                    # 正则匹配 language: path
                    match = re.match(r"(.*):\s*(.*)", line)
                    if not match:
                        raise Exception(
                            f"File [{path}]\n\tline [{line_num}]: Package definition error"
                        )
                    # 检查 package language 是否已定义
                    for package in arpc_meta['package']:
                        if package['language'] == match.group(1):
                            raise Exception(
                                f"File [{path}]\n\tline [{line_num}]: Repeated package: {match.group(1)}"
                            )
                    # 保存 package
                    arpc_meta['package'].append({
                        'language': match.group(1),
                        'name': match.group(2),
                        'path': match.group(2)
                    })
            # 正在处理 procedure
            elif handle_procedures:
                if line.startswith("}"):
                    handle_procedures = False
                    continue
                elif line.startswith("{"):
                    continue
                else:
                    # 正则匹配 procedure name(request): response
                    match = re.match(
                        r"procedure\s+(\w+)\s*\((\w+)\):\s*(\w+)", line)
                    if not match:
                        raise Exception(
                            f"File [{path}]\n\tline [{line_num}]: Procedure definition error"
                        )
                    # 检查 procedure 名称是否重复
                    for procedure in arpc_meta['procedures']:
                        if procedure['name'] == match.group(1):
                            raise Exception(
                                f"File [{path}]\n\tline [{line_num}]: Repeated procedure: {match.group(1)}"
                            )
                    # 保存 procedure
                    arpc_meta['procedures'].append({
                        'name': match.group(1),
                        'index': len(arpc_meta['procedures']),
                        'request': match.group(2),
                        'response': match.group(3)
                    })
                continue
            # 正在处理 param
            elif handle_param:
                if line.startswith("}"):
                    handle_param = False
                    continue
                elif line.startswith("{"):
                    continue
                else:
                    # 正则匹配 name: type = index
                    match = re.match(r"\s*(\w+):\s*(\w+)\s*=\s*(\d+)", line)
                    if not match:
                        raise Exception(
                            f"File [{path}]\n\tline [{line_num}]: Param definition error"
                        )
                    index = int(match.group(3))
                    # 检查 index 是否重复
                    for param in arpc_meta['param'][param_name]:
                        if param['index'] == index:
                            raise Exception(
                                f"File [{path}]\n\tline [{line_num}]: Repeated param index: {index}"
                            )
                    arpc_meta['param'][param_name].append({
                        'name': match.group(1),
                        'type': match.group(2),
                        'index': index
                    })
                continue
            else:
                if line.startswith("package"):
                    if 'package' in arpc_meta:
                        raise Exception(
                            f"File [{path}]\n\tline [{line_num}]: Repeated package area"
                        )
                    handle_package = True
                    arpc_meta['package'] = []  # 保存 package
                    continue
                elif line.startswith("procedures"):
                    if 'procedures' in arpc_meta:
                        raise Exception(
                            f"File [{path}]\n\tline [{line_num}]: Repeated procedures area"
                        )
                    handle_procedures = True
                    arpc_meta['procedures'] = []  # 保存 procedure
                    continue
                elif line.startswith("param"):
                    # 正则匹配 param RequestV1 {
                    match = re.match(r"param\s+(\w+)\s+{", line)
                    if not match:
                        raise Exception(
                            f"File [{path}]\n\tline [{line_num}]: Param name definition error"
                        )
                    # 获取参数名
                    param_name = match.group(1)
                    if param_name in arpc_meta:
                        raise Exception(
                            f"File [{path}]\n\tline [{line_num}]: Repeated param: {param_name}"
                        )
                    handle_param = True
                    if 'param' not in arpc_meta:
                        arpc_meta['param'] = {}
                    arpc_meta['param'][param_name] = []
                elif line.startswith("arpc"):
                    # 正则匹配 "arpc: *"
                    result = re.match(r"arpc:\s*(.*)", line)
                    if result:
                        if 'version' in arpc_meta:
                            raise Exception(
                                f"File [{path}]\n\tline [{line_num}]: Repeated arpc line"
                            )
                        arpc_meta['version'] = result.group(1)
                elif line.startswith("unique"):
                    # 正则匹配 "arpc: *"
                    result = re.match(r"unique:\s*(.*)", line)
                    if result:
                        if 'unique' in arpc_meta:
                            raise Exception(
                                f"File [{path}]\n\tline [{line_num}]: Repeated arpc line"
                            )
                        arpc_meta['unique'] = result.group(1)
                else:
                    pass
    return arpc_meta


def run(path: str | List[str]):
    if isinstance(path, str):
        arpc = compile_arpc(path)
        result = generate_package(arpc, path)
        return result
    else:
        result = []
        for p in path:
            arpc = compile_arpc(p)
            result.append(generate_package(arpc, p))
        return result


def compile_(input_path: str, output_path: str, async_: bool = False):
    arpc = compile_arpc(input_path)
    generate_package(arpc, input_path, output_path, async_)


def compiles(input_path: str, output_path: str, async_: bool = False):
    files = []
    for root, _, fs in os.walk(input_path):
        for f in fs:
            tmp_path = os.path.join(root, f)
            if os.path.isfile(tmp_path) and f.endswith(".arpc"):
                files.append(tmp_path)
    for file_ in files:
        print(f"Compiling{' Async: ' if async_ else ': '}{file_}")
        c_path = os.path.relpath(file_, input_path)
        generate = os.path.dirname(os.path.join(output_path, c_path))
        compile_(file_, generate, async_)
